import streamlit as st
from streamlit import session_state as st_state
import pyfurnace as pf
from .motif_command import MotifCommand


class StemCommand(MotifCommand):

    def execute(self, motif=None):
        ### Modify the motif
        if motif:
            # interface data:
            # top_seq, seq_length, wobble_interval, wobble_tolerance
            data = self.interface(
                "mod",
                motif[0].sequence,
                motif.length,
                motif.wobble_interval,
                motif.wobble_tolerance,
            )
            top_seq, seq_length, wobble_interval, wobble_tolerance = data

            if top_seq and motif[0].sequence != top_seq:
                st_state.modified_motif_text += (
                    f"\nmotif.set_up_sequence(" f"'{top_seq}')"
                )
                motif.set_up_sequence(top_seq)

            elif seq_length and motif.length != seq_length:
                st_state.modified_motif_text += f"\nmotif.length = {seq_length}"
                motif.length = seq_length

            elif motif.wobble_interval != wobble_interval:
                st_state.modified_motif_text += (
                    f"\nmotif.wobble_interval = " f"{wobble_interval}"
                )
                motif.wobble_interval = wobble_interval

            elif motif.wobble_tolerance != wobble_tolerance:
                st_state.modified_motif_text += (
                    f"\nmotif.wobble_tolerance = " f"{wobble_tolerance}"
                )
                motif.wobble_tolerance = wobble_tolerance

        ### Create a new motif
        else:
            top_seq, seq_length, wobble_interval, wobble_tolerance = self.interface()
            if top_seq:
                st_state.motif_buffer = f"motif = pf.Stem(sequence = '{top_seq}')"
                motif = pf.Stem(sequence=top_seq)

            else:
                st_state.motif_buffer = (
                    f"motif = pf.Stem(length = {seq_length}, "
                    f"wobble_interval = {wobble_interval}, "
                    f"wobble_tolerance = {wobble_tolerance}, "
                    f"wobble_insert = 'middle', "
                    f"strong_bases = True)"
                )
                motif = pf.Stem(
                    length=seq_length,
                    wobble_interval=wobble_interval,
                    wobble_tolerance=wobble_tolerance,
                )

            # save the motif in the session state
            st_state.motif = motif

    def interface(
        self, key="", top_seq=None, len_default=7, wobble_interval=5, wobble_tolerance=2
    ):

        ### initialize the variables
        seq_length = 0

        ### create the interface
        col1, col2 = st.columns([1, 5], vertical_alignment="bottom")

        with col1:
            specific_seq = st.toggle("Custom Sequence", key=f"seq_stem{key}")

        with col2:
            if specific_seq:
                col1, col2 = st.columns([5, 1])
                with col1:
                    top_seq = st.text_input(
                        "Sequence:", key=f"txt_top_seq_stem{key}", value=top_seq
                    )
            else:
                subcol1, subcol2, subcol3 = st.columns([3, 1, 1])
                with subcol1:
                    seq_length = st.number_input(
                        "Length:",
                        key=f"stem_length{key}",
                        min_value=1,
                        value=len_default,
                    )
                with subcol2:
                    wobble_interval = st.number_input(
                        "Wobble interval:",
                        key=f"stem_wobble_interv{key}",
                        min_value=0,
                        value=wobble_interval,
                        help="Add a wobble every n° " "nucleotides",
                    )
                with subcol3:
                    wobble_tolerance = st.number_input(
                        "Wobble tolerance:",
                        key=f"stem_wobble_tole{key}",
                        min_value=0,
                        value=wobble_tolerance,
                    )

        return top_seq, seq_length, wobble_interval, wobble_tolerance



import streamlit as st
from streamlit import session_state as st_state
import pyfurnace as pf
from .motif_command import MotifCommand


class TetraLoopCommand(MotifCommand):

    def execute(self, motif=None):
        ### Modify the motif
        if motif:
            seq, flip = self.interface("mod", motif[0].sequence)
            if seq and motif[0].sequence != seq:
                st_state.modified_motif_text += f"\nmotif.set_sequence('{seq}')"
                motif.set_sequence(seq)
            elif flip:
                st_state.modified_motif_text += "\nmotif.flip()"
                motif.flip()

        ### Create a new motif
        else:
            seq, open_left = self.interface()
            if seq:
                st_state.motif_buffer = (
                    f"motif = pf.TetraLoop({open_left}, " f"sequence = '{seq}')"
                )
                motif = pf.TetraLoop(open_left, sequence=seq)

            else:
                st_state.motif_buffer = f"motif = pf.TetraLoop({open_left})"
                motif = pf.TetraLoop(open_left)
            # save the motif in the session state
            st_state.motif = motif

    def interface(self, key="", seq_default="UUCG"):
        seq = None
        open_left = False
        col1, col2, col3 = st.columns([1, 1, 4], vertical_alignment="bottom")
        with col1:
            custom_seq = st.toggle("Custom Sequence", key=f"seq_tetra{key}")
        with col2:
            if key == "mod":
                open_left = st.button("Flip", key=f"open_left_tetra{key}")
            else:
                open_left = st.toggle(
                    "Flip",
                    value=st_state.current_line_occupied,
                    key=f"open_left_tetra{key}",
                )
        if custom_seq:
            with col3:
                seq = st.text_input(
                    "Sequence:", value=seq_default, key=f"txt_seq_tetra{key}"
                )
        return seq, open_left



import streamlit as st
from streamlit import session_state as st_state
from streamlit_option_menu import option_menu
import re
from Bio.Seq import Seq
from Bio.SeqUtils import MeltingTemp as mt
from Bio.SeqUtils import gc_fraction, molecular_weight
import warnings
import pyfurnace as pf
from utils import load_logo, main_menu_style, copy_to_clipboard, write_format_text
from utils.template_functions import symbols, check_dimer, sanitize_input, reference


def initialize_session_state():
    params_value = {"rna_origami_seq": "", "promoter": pf.T7_PROMOTER}
    for key, value in params_value.items():
        if key not in st_state:
            st_state[key] = value


def check_seq():
    # take the input sequence and sanitize it
    seq = sanitize_input(
        st.text_input("Input sequence (DNA or RNA):", value=st_state["rna_origami_seq"])
    )
    # check the symbols in the sequence
    if set(seq) - symbols:
        st.warning(
            "The sequence contains symbols not included in the "
            "[IUPAC alphabet](https://www.bioinformatics.org/sms/iupac.html).",
            icon="⚠️",
        )

    if seq != st_state["rna_origami_seq"]:
        st_state["rna_origami_seq"] = seq
        st.rerun()

    seq = Seq(seq)
    return seq


def rna_to_template():
    col1, col2 = st.columns([1, 5])
    # st.write("#### DNA template:")
    promoter = st.text_input(
        "**Add a promoter** (default: T7 promoter)", value=st_state.promoter
    )

    # take a specific promoter sequence for the DNA template
    # and check that the RNA sequence starts with G
    if promoter == pf.T7_PROMOTER and seq[0] != "G":
        st.warning("The RNA sequence doesn't start with G", icon="⚠️")
    elif promoter != st_state.promoter:
        st_state.promoter = promoter
        st.rerun()

    dna_template = Seq(promoter) + seq.back_transcribe()
    non_coding = dna_template.reverse_complement()
    col1, col2 = st.columns([1, 5])

    # coding strand
    copy_ratio = [1, 8]
    col1, col2 = st.columns(copy_ratio, vertical_alignment="center")
    with col1:
        copy_to_clipboard(dna_template, "Coding strand:")
    with col2:
        write_format_text(dna_template)
    col1, col2 = st.columns(copy_ratio, vertical_alignment="center")
    with col1:
        copy_to_clipboard(non_coding, "Non-coding strand:")
    with col2:
        write_format_text(non_coding)

    # Save the DNA template in the session state and add a link to the primer page
    st_state["dna_template"] = str(dna_template)
    st_state.prepare_ind = 0
    st.page_link(
        "pages/4_Prepare.py",
        label=":orange[Prepare the PCR primers]",
        icon=":material/sync_alt:",
    )


def convert_tab(seq):
    """
    Calculate the main property of the sequence (gc content, molecular weight,
    nucleotide composition, melting temperature) and convert the sequence to the
    different formats: reverse, complement, reverse complement, RNA transcribed,
    DNA template.
    """
    # format the input sequence
    # st.write(f"Your sequence ({len(seq)} nt):")
    st.write(f"Sequence type: {seq_type}. Length: {len(seq)} bases")
    # write_format_text(seq)

    # calculate the main properties of the sequence
    col1, col2, col3, col4 = st.columns(4)
    with col1:
        st.write("GC content (%)")
        write_format_text(round(gc_fraction(seq, ambiguous="ignore") * 100, 2))
    with col2:
        st.write("Molecular weight (Dalton)")
        write_format_text(f"{molecular_weight(seq, seq_type):.3e}")
    with col3:
        st.write("Nucleotide composition")
        bases = set(str(seq))
        t = ""
        for b in bases:
            t = t + f"{b}: {str(seq).count(b)}; "
        write_format_text(t)
    with col4:
        st.write("Melting temperature (°C)")
        if seq_type == "RNA":
            nn_table = mt.RNA_NN3
            complement = seq.complement_rna()
        else:
            nn_table = mt.DNA_NN4
            complement = seq.complement()
        write_format_text(round(mt.Tm_NN(seq, c_seq=complement, nn_table=nn_table), 2))

    # Useful conversions
    col1, col2, col3 = st.columns(3)
    with col1:
        copy_to_clipboard(seq[::-1], "Reverse:")
        write_format_text(seq[::-1])
    with col2:
        copy_to_clipboard(complement, "Complement:")
        write_format_text(complement)
    with col3:
        if seq_type == "DNA":
            rev_complement = seq.reverse_complement()
        else:
            rev_complement = seq.reverse_complement_rna()
        copy_to_clipboard(rev_complement, "Reverse complement:")
        write_format_text(rev_complement)

    if seq_type == "DNA":
        # coding strand
        st.write("**DNA to RNA:**")
        col1, col2 = st.columns([1, 8], vertical_alignment="center")
        rna = seq.transcribe()
        with col1:
            copy_to_clipboard(rna, "RNA")
        with col2:
            write_format_text(rna)

    # if the sequence is RNA, create a DNA template
    else:
        rna_to_template()


def search_tab(seq):
    """Search a subsequence and highlight the bases"""
    subseq_list = []
    i = 0
    # take a subsequence
    subseq = sanitize_input(st.text_input("Search for the subsequence:", key=f"seq{i}"))
    subseq_list.append(subseq)
    while subseq_list[-1]:
        # take the last subsequence
        sub_seq = subseq_list[-1]
        # calculate the indexes of the subsequence in the sequence
        indexes = [substr.start() + 1 for substr in re.finditer(str(sub_seq), str(seq))]
        if not indexes:
            st.error("Subsequence not found", icon=":material/personal_injury:")
        else:
            # highlight in red the subsequence in the sequence
            highlighted = str(seq).replace(sub_seq, f":red[{sub_seq}]")
            st.write(f"Sequence found at index: {indexes}")
            st.markdown(highlighted)
        # add the space for another subsequence
        st.divider()
        i += 1  # add another subsequence with a different widget key
        subseq = sanitize_input(
            st.text_input("Search for the subsequence:", key=f"seq{i}")
        )
        # add the subsequence to the subsequence list
        subseq_list.append(subseq)
        # if the last subsequences is empty, stop the loop


def dimer_tab(seq):
    """Check the dimer between the main sequence and a list of subsequences
    and highlight the aligned bases"""
    subseq_list = []
    i = 0
    # take a subsequence
    subseq = sanitize_input(
        st.text_input("Check dimer with the sequence:", key=f"dim{i}")
    )
    subseq_list.append(subseq)
    while subseq_list[-1]:
        # take the last subsequence
        sub_seq = subseq_list[-1]
        st.write("Best dimer found:")
        # calculate the best dimer between the sequence and the subsequence
        dimer_subseq = check_dimer(seq, sub_seq, basepair=basepair)
        # write the dimer that is formed
        write_format_text(dimer_subseq)
        # add expander with all dimers found for the subsequence
        with st.expander("All dimers"):
            # calculate all the dimers between the sequence and the subsequence
            dimer_dict = check_dimer(seq, sub_seq, dict_format=True, basepair=basepair)
            # add a slider to select the number of basepairs
            n_pairs = st.slider(
                "Dimer with n° of basepairs:",
                min_value=min(dimer_dict),
                max_value=max(dimer_dict),
                value=max(dimer_dict),
                key=f"dimers{i}",
            )
            if n_pairs in dimer_dict:
                for other_dimer in dimer_dict[n_pairs]:
                    write_format_text(other_dimer)
            else:
                st.write(f"No Dimer found for {n_pairs} basepairs")
        # add the space for another subsequence
        st.divider()
        i += 1
        subseq = sanitize_input(
            st.text_input("Search for the subsequence:", key=f"dim{i}")
        )
        subseq_list.append(subseq)
        # if the last subsequences is empty, stop the loop


if __name__ == "__main__":
    ### set the logo of the app
    load_logo()
    warnings.filterwarnings("ignore")  # ignore warnings

    initialize_session_state()

    st.header(
        "Convert",
        help="Prepare the DNA template for you RNA Origami, "
        "align structures and search for dimers.",
    )

    seq = check_seq()

    seq_type = "RNA"
    if "T" in seq and "U" in seq:
        st.error(
            "Both T and U found in the sequence", icon=":material/personal_injury:"
        )
    elif "T" in seq:
        seq_type = "DNA"
    elif "T" not in seq and "U" not in seq:
        seq_type = st.radio(
            "Choose the sequence type: (usually auto-detected)",
            ["DNA", "RNA"],
            horizontal=True,
        )

    # according to the sequence type, create the basepair dictionary
    basepair = str.maketrans("UACG", "AUGC")  # RNA basepair
    if seq_type == "DNA":
        basepair = str.maketrans("TACG", "ATGC")

    # create the tabs with the functions
    st.write("\n")  # add space between initial menu and motif menu
    option_data = {
        "Convert": "bi bi-arrow-repeat",
        "Search": "bi bi-align-center",
        "Dimer": "bi bi-bar-chart-steps",
    }

    selected_operation = option_menu(
        None,
        list(option_data.keys()),
        icons=list(option_data.values()),
        menu_icon="cast",
        orientation="horizontal",
        styles=main_menu_style,
    )

    if not seq:
        st.stop()
    elif selected_operation == "Convert":
        convert_tab(seq)
    elif selected_operation == "Search":
        search_tab(seq)
    elif selected_operation == "Dimer":
        dimer_tab(seq)

    # add bibliography
    reference()



import inspect
from pyfurnace.design import utils, start_end_stem

import streamlit as st
from streamlit_option_menu import option_menu
from .motif_command import MotifCommand
from .general_edit_command import GeneralEditCommand
from .tetraloop_command import TetraLoopCommand
from .. import second_menu_style
from ..motifs_icons import MOTIF_ICONS

# Filter and collect the motif utils
util_names = [
    ut_name
    for ut_name, obj in inspect.getmembers(utils.motif_lib)
    if inspect.isfunction(obj)
]

# ignore the simple vertical connection
del util_names[util_names.index("vertical_link")]

# Add the tetraloop option
util_names = ["Tetraloop"] + util_names


class ConnectionsCommand(MotifCommand):

    def execute(self, motif=None):
        ### Modify motif
        if motif:
            GeneralEditCommand().execute(motif)

        ### Create new motif
        else:
            util_option = option_menu(
                None,
                util_names,
                icons=[MOTIF_ICONS[name] for name in util_names],
                menu_icon="cast",
                orientation="horizontal",
                styles=second_menu_style,
                key="UtilsOption",
            )

            if util_option == "Tetraloop":
                TetraLoopCommand().execute()
                return
            if util_option == "start_end_stem":
                start_end_stemCommand().execute()
                return

            name = util_option.replace(" ", "_")
            motif_util = getattr(utils.motif_lib, name)
            motif_text = f"motif = pf.{name}"
            flip_vert, flip_hor, rotate = self.interface()
            if flip_vert or flip_hor:
                motif_text += f".flip({flip_vert}, {flip_hor})"
            if rotate:
                motif_text += f".rotate({rotate})"
            # save the motif in the session state
            st.session_state.motif_buffer = (
                f"motif = pf.{name}(hflip={flip_hor}, "
                f"vflip={flip_vert}, rotate={rotate})"
            )
            st.session_state.motif = motif_util(
                hflip=flip_hor, vflip=flip_vert, rotate=rotate
            )

    def interface(self, key=""):
        return GeneralEditCommand.interface(key)


class start_end_stemCommand(MotifCommand):

    def execute(self, motif=None):
        t_l, t_r, b_l, b_r = self.interface()
        st.session_state.motif_buffer = (
            f"motif = pf.start_end_stem(up_left='{t_l}', "
            f"up_right='{t_r}', down_left='{b_l}', "
            f"down_right='{b_r}')"
        )
        st.session_state.motif = start_end_stem(
            up_left=t_l, up_right=t_r, down_left=b_l, down_right=b_r
        )

    def interface(
        self,
        key="",
        up_l_def="3",
        up_r_def="5",
        down_l_def="5",
        down_r_def="3",
        up_ind=0,
        down_ind=1,
    ):
        _, _, col1, col2, _, _ = st.columns([1, 1, 1, 1, 1, 1])
        with col1:
            t_l = st.selectbox(
                "Top left:",
                [up_l_def] + ["─", None],
                index=up_ind,
                key=f"start_end_stem_up_left{key}",
            )
            b_l = st.selectbox(
                "Bottom left:",
                [down_l_def] + ["─", None],
                index=down_ind,
                key=f"start_end_stem_down_left{key}",
            )
        with col2:
            t_r = st.selectbox(
                "Top right:",
                [up_r_def] + ["─", None],
                index=up_ind,
                key=f"start_end_stem_up_right{key}",
            )
            b_r = st.selectbox(
                "Bottom right:",
                [down_r_def] + ["─", None],
                index=down_ind,
                key=f"start_end_stem_down_right{key}",
            )
        if not t_l:
            t_l = ""
        if not t_r:
            t_r = ""
        if not b_l:
            b_l = ""
        if not b_r:
            b_r = ""
        return t_l, t_r, b_l, b_r
